; temporr die PATH-Variable erweitern (damit alle dlls gefunden werden)
EnvSet("PATH", EnvGet("PATH") & ";" & @ScriptDir & (@AutoItX64 ? "\x64\" : "\x86\"))

Global Const $h_KRNL32 = DllOpen("kernel32.dll")
Global Const $p_KRNL32 = DllCall($h_KRNL32, "handle", "GetModuleHandleW", "wstr", "kernel32.dll")[0]
Global Const $p_PROCHEAP = DllCall($h_KRNL32, "HANDLE", "GetProcessHeap")[0] ; Pointer to the process heap (for HeapAlloc/HeapFree)
Global Const $p_FLANNDLL = DllCall($h_KRNL32, "handle", "LoadLibraryW", "wstr", "flann.dll")[0]

; Struktur FLANNPARAMETERS
Global Const $tag_FLANNParameters = "struct;INT algorithm;INT checks;FLOAT cb_index;FLOAT eps;INT trees;INT leaf_max_size;INT branching;INT iterations;INT centers_init;FLOAT target_precision;FLOAT build_weight;FLOAT memory_weight;FLOAT sample_fraction;UINT table_number_;UINT key_size_;UINT multi_probe_level_;INT log_level;LONG random_seed;endstruct"
Global $p_DEFAULT_FLANN_PARAMETERS = DllStructCreate($tag_FLANNParameters)
$p_DEFAULT_FLANN_PARAMETERS.algorithm = 1 ; = FLANN_INDEX_KDTREE
$p_DEFAULT_FLANN_PARAMETERS.checks = 32
$p_DEFAULT_FLANN_PARAMETERS.cb_index = 0.2
$p_DEFAULT_FLANN_PARAMETERS.eps = 0.0
$p_DEFAULT_FLANN_PARAMETERS.trees = 4
$p_DEFAULT_FLANN_PARAMETERS.leaf_max_size = 4
$p_DEFAULT_FLANN_PARAMETERS.branching = 32
$p_DEFAULT_FLANN_PARAMETERS.iterations = 11
$p_DEFAULT_FLANN_PARAMETERS.centers_init = 0 ; = FLANN_CENTERS_RANDOM
$p_DEFAULT_FLANN_PARAMETERS.target_precision = 0.9
$p_DEFAULT_FLANN_PARAMETERS.build_weight = 0.01
$p_DEFAULT_FLANN_PARAMETERS.memory_weight = 0.0
$p_DEFAULT_FLANN_PARAMETERS.sample_fraction = 0.1
$p_DEFAULT_FLANN_PARAMETERS.log_level = 0 ; = FLANN_LOG_NONE
$p_DEFAULT_FLANN_PARAMETERS.random_seed = 0

; Distance-Type festlegen (hier euklidische):
DllCall("flann.dll", "VOID", "flann_set_distance_type", "INT", 1, "INT", 1)

Func _flann_find_nearest_neighbors($t_DataSet, $t_TestSet, $nn = 1)
	; By AspirinJunkie
	Local Static $p_A = DllCall($h_KRNL32, "ptr", "GetProcAddress", "handle", $p_FLANNDLL, "str", "flann_find_nearest_neighbors_double")[0]
	Local $p_indices = DllStructCreate("INT[" & $t_TestSet[1] * $nn & "]")
	Local $p_dists = DllStructCreate("DOUBLE[" & $t_TestSet[1] * $nn & "]")

	DllCallAddress("INT", $p_A, _
			"PTR", $t_DataSet[0], "INT", $t_DataSet[1], "INT", $t_DataSet[2], _
			"PTR", $t_TestSet[0], "INT", $t_TestSet[1], _
			"PTR", DllStructGetPtr($p_indices), "PTR", DllStructGetPtr($p_dists), _
			"INT", $nn, "Ptr", DllStructGetPtr($p_DEFAULT_FLANN_PARAMETERS))

	Local $a_Indices = _FLANN_Data2Au3(DllStructGetPtr($p_indices), $t_TestSet[1], $nn, "INT")
	Local $a_dists = _FLANN_Data2Au3(DllStructGetPtr($p_dists), $t_TestSet[1], $nn)
	Local $a_Ret[2] = [$a_Indices, $a_dists]
	Return $a_Ret
EndFunc   ;==>_flann_find_nearest_neighbors



Func _FLANN_Data2Au3(ByRef Const $V, $r = 1, $c = 1, $Type = "DOUBLE")
	If IsArray($V) Then
		Local $pt = $V[0]
		$r = $V[1]
		$c = $V[2]
	Else
		Local $pt = $V
	EndIf
	Local $N = Int($r * $c)

	$strct = DllStructCreate($Type & "[" & $N & "]", $pt)
	Local $a_Ret[$r][$c]

	; save whole matrix in row-major-order (C-order)
	Local $x = 1
	For $i = 0 To $r - 1
		For $j = 0 To $c - 1
			$a_Ret[$i][$j] = DllStructGetData($strct, 1, $x)
			$x += 1
		Next
	Next
	Return $a_Ret
EndFunc   ;==>_FLANN_Data2Au3


Func _FLANN_Data2C(ByRef Const $V)
	; By AspirinJunkie
	Local Static $p_A = DllCall($h_KRNL32, "ptr", "GetProcAddress", "handle", $p_KRNL32, "str", "HeapAlloc")[0] ; get the function address of HeapAlloc (for memory allocation)
	Local $r = UBound($V, 1)
	Local $c = UBound($V, 2)
	Local $N = Int($r * $c)

	; neuer Speicher im Heap
	Local $h_Pt = DllCallAddress("PTR", $p_A, "HANDLE", $p_PROCHEAP, "DWORD", 0x1, "ulong_ptr", 8 * $N)[0] ; 4 Byte, da Float-Werte
	; Neue Struktur auf dem allozierten Speicher erstellen
	Local $strct = DllStructCreate("DOUBLE [" & $N & "]", $h_Pt)

	; save whole matrix in row-major-order (C-order)
	Local $x = 1
	For $i = 0 To $r - 1
		For $j = 0 To $c - 1
			DllStructSetData($strct, 1, Number($V[$i][$j], 3), $x)
			$x += 1
		Next
	Next

	Local $a_Ret[3] = [$h_Pt, $r, $c]
	Return $a_Ret
EndFunc   ;==>_FLANN_Data2C



Func _FLANN_DelMatrix(ByRef $V)
	; By AspirinJunkie
	Local Static $p_F = DllCall($h_KRNL32, "ptr", "GetProcAddress", "handle", $p_KRNL32, "str", "HeapFree")[0]
	DllCallAddress("BOOL", $p_F, "handle", $p_PROCHEAP, "DWORD", 1, "ptr", $V[0])
	If @error Then Return SetError(1, @extended, False)
	$V = ""
	Return True
EndFunc   ;==>_FLANN_DelMatrix

